require 'rack'
require 'active_support/hash_with_indifferent_access'
require 'active_support/core_ext/object/deep_dup'
require 'pathname'
require 'generic/target'
require 'generic/routing'
require 'slim'

module Generic

  class Request < Rack::Request
    def params
      @params ||= ActiveSupport::HashWithIndifferentAccess.new(super)
    end
  end

  class InsEval
    def self.instance(scope, &block)
      scope.instance_eval &block
    end

    def self.apply(scope, &handler)
      scope.apply &handler
    end

    def self.hand(scope_class, handler, before_actions, after_actions)
      Proc.new do |env|
        scope = scope_class.new(env)
        response = catch (:halt) do
          before_actions.each do |action|
            Generic::InsEval.instance(scope, &action)
          end
          Generic::InsEval.apply(scope, &handler)
        end
        catch (:halt) do
          after_actions.each do |action|
            Generic::InsEval.apply(scope, &action)
          end
        end
        response.finish
      end
    end

  end

  class Response < Rack::Response
  end

  def self.root
    @root ||= Pathname.pwd
  end

  class Base
    include Genericer::General

    # General
    general(:scope, Class.new(Target))
    general(:rd, [])
    general(:before_actions, [])
    general(:after_actions, [])
    general(:guard, {})
    general(:routes, {})
    general(:impt, [])

    def initialize(app=nil)
      # Rack Builder
      builder = Rack::Builder.new

      #Register Importing Classes
      self.class.impt.each do |im, args, block|
        builder.use(im, *args, &block)
      end

      # Register group nested
      self.class.routes.each do |url, a_class|
        builder.map(url) do
          use a_class
        end
      end

      builder.run Generic::Routing.new({
                                           :scope => self.class.scope,
                                           :rd => self.class.rd,
                                           :before_actions => self.class.before_actions,
                                           :after_actions => self.class.after_actions,
                                           :fail => app
                                       })

      @app = builder.to_app
    end


    def call(env)
      @app.call(env)
    end

    def settings
      self.class.settings
    end

    class << self

      # Methods

      def get(path, options={}, &block)
        options[:constraints] = guard.merge(options[:constraints] || {})
        options[:constraints].merge!(:request_method => "GET")
        define_route(path, options, &block)
      end

      def post(path, options={}, &block)
        options[:constraints] = guard.merge(options[:constraints] || {})
        options[:constraints].merge!(:request_method => "POST")
        define_route(path, options, &block)
      end

      def put(path, options={}, &block)
        options[:constraints] = guard.merge(options[:constraints] || {})
        options[:constraints].merge!(:request_method => "PUT")
        define_route(path, options, &block)
      end

      def delete(path, options={}, &block)
        options[:constraints] = guard.merge(options[:constraints] || {})
        options[:constraints].merge!(:request_method => "DELETE")
        define_route(path, options, &block)
      end

      # Group Route
      def group(url, app=nil, &block)
        scope = self.scope
        app = Class.new(self.superclass) do
          self.scope = scope
          class_eval(&block)
        end
        routes[url] = app
      end

      # Register Extension
      def register(*extensions)
        extensions.each do |ext|
          extend ext
          if ext.check(self)
            ext.respond_to?(:registered)
          end
        end
      end

      # Set Guard
      def set_guard(args, &block)
        current = self.guard.dup
        self.guard = args
        instance_eval(&block)
        self.guard = current
      end

      # System Helper
      def library(*args, &block)
        if block_given?
          args << Module.new(&block)
        end

        args.each do |m|
          scope.send :include, m
        end
      end

      # Import
      def import(cls, *args, &block)
        impt << [cls, args, block]
      end

      # Before Proccess
      def before(&block)
        before_actions << Proc.new(&block)
      end

      # After Proccess
      def after(&block)
        after_actions << Proc.new(&block)
      end
      
      protected
      # Define Framework Route
      def define_route(path, options, &block)
        self.rd << [path, options, Proc.new(&block)]
      end
    end
  end
end
