require 'forwardable'
require 'rack/contrib/cookies'

module Generic
  class Target
    extend Forwardable

    attr_reader :request, :response, :attr_reader
    def_delegators :request, :session, :params
    def_delegators :response, :headers

    def initialize(p)
      @request = Request.new(p)
      @response = Response.new [], 200, {'Content-Type' => 'text/html'}
    end

    private
    def system_redirect(uri, status=302)
      exit(status, {'Location' => uri})
    end

    def status(code)
      response.status = code
    end

    def exit(status, headers={}, body='')
      @response = Response.new(body, status, headers)
      cookies.finish!(response) if @cookies
      throw :halt, response
    end

    def system_cookies
      @cookies ||= Rack::Cookies::CookieJar.new(request.cookies)
    end
    [:goto, :go_to].each do |keyword|
      alias_method keyword, :system_redirect
    end

    public
    def apply(&handler)
      data = instance_eval(&handler)
      data.respond_to?(:each) ? response.body = data : response.write(data)
      cookies.finish!(response) if @cookies
      response
    end
  end
end
